﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.ConditionActionEncoding {
  [Item("ActionSetSubsumptionoperator", "Description missing")]
  [StorableClass]
  public class ActionSetSubsumptionoperator : SingleSuccessorOperator, IActionSetSubsumption {

    #region Parameter Properties
    public ILookupParameter<ItemArray<IClassifier>> ClassifiersParameter {
      get { return (ILookupParameter<ItemArray<IClassifier>>)Parameters["Classifiers"]; }
    }
    public ILookupParameter<ItemArray<IntValue>> NumerositiesParameter {
      get { return (ILookupParameter<ItemArray<IntValue>>)Parameters["Numerosities"]; }
    }
    public ILookupParameter<ItemArray<IntValue>> ExperiencesParameter {
      get { return (ILookupParameter<ItemArray<IntValue>>)Parameters["Experiences"]; }
    }
    public ILookupParameter<ItemArray<DoubleValue>> ErrorsParameter {
      get { return (ILookupParameter<ItemArray<DoubleValue>>)Parameters["Errors"]; }
    }
    public ILookupParameter<DoubleValue> ErrorZeroParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["ErrorZero"]; }
    }
    public ILookupParameter<IntValue> ThetaSubsumptionParameter {
      get { return (ILookupParameter<IntValue>)Parameters["ThetaSubsumption"]; }
    }
    protected ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }
    public IScope CurrentScope {
      get { return CurrentScopeParameter.ActualValue; }
    }

    public string HasBeenSubsumedParameterName {
      get { return "HasBeenSubsumed"; }
    }
    #endregion

    [StorableConstructor]
    protected ActionSetSubsumptionoperator(bool deserializing) : base(deserializing) { }
    protected ActionSetSubsumptionoperator(ActionSetSubsumptionoperator original, Cloner cloner)
      : base(original, cloner) {
    }
    public ActionSetSubsumptionoperator()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<IClassifier>("Classifiers"));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("Numerosities"));
      Parameters.Add(new ScopeTreeLookupParameter<IntValue>("Experiences"));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Errors"));
      Parameters.Add(new LookupParameter<DoubleValue>("ErrorZero"));
      Parameters.Add(new LookupParameter<IntValue>("ThetaSubsumption"));
      Parameters.Add(new ScopeParameter("CurrentScope"));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new ActionSetSubsumptionoperator(this, cloner);
    }

    public override IOperation Apply() {
      var subscopes = CurrentScope.SubScopes;
      var classifiers = ClassifiersParameter.ActualValue;
      var numerosities = NumerositiesParameter.ActualValue;
      var experiences = ExperiencesParameter.ActualValue;
      var errors = ErrorsParameter.ActualValue;
      if (subscopes.Count != classifiers.Length || subscopes.Count != numerosities.Length
        || subscopes.Count != experiences.Length || subscopes.Count != errors.Length) {
        throw new ArgumentException("The number of subscopes is not equal to the number of classifiers.");
      }

      int subsumptionClassifier = -1;

      for (int i = 0; i < classifiers.Length; i++) {
        if (CouldSubsume(experiences[i], errors[i])) {
          if (subsumptionClassifier == -1 || classifiers[i].IsMoreGeneral(classifiers[subsumptionClassifier])) {
            subsumptionClassifier = i;
          }
        }
      }

      if (subsumptionClassifier != -1) {
        for (int i = 0; i < classifiers.Length; i++) {
          if (classifiers[subsumptionClassifier].IsMoreGeneral(classifiers[i])) {
            numerosities[subsumptionClassifier].Value += numerosities[i].Value;
            AddVariableToScope(subscopes[i], new BoolValue(true));
          } else {
            AddVariableToScope(subscopes[i], new BoolValue(false));
          }
        }
      } else {
        for (int i = 0; i < classifiers.Length; i++) {
          AddVariableToScope(subscopes[i], new BoolValue(false));
        }
      }

      return base.Apply();
    }

    private void AddVariableToScope(IScope scope, BoolValue boolValue) {
      if (scope.Variables.ContainsKey(HasBeenSubsumedParameterName)) {
        scope.Variables[HasBeenSubsumedParameterName].Value = boolValue;
      } else {
        scope.Variables.Add(new Variable(HasBeenSubsumedParameterName, boolValue));
      }
    }

    private bool CouldSubsume(IntValue exp, DoubleValue error) {
      return exp.Value > ThetaSubsumptionParameter.ActualValue.Value && error.Value < ErrorZeroParameter.ActualValue.Value;
    }
  }
}
